/* Optimized (Intel Intrinsics) implementation of SpoC-64 AEAD
   Written by:
   Kalikinkar Mandal <kmandal@uwaterloo.ca>/<kmandal@unb.ca>
   Date: August 21, 2020
   This implementation supports when message or ad lengths are a multiple of 8 bits.
*/

#include "spoc64.h"
static inline __m256i _mm256_loadu2_m128i(__m128i *low,__m128i *high)
{
	return _mm256_inserti128_si256(_mm256_castsi128_si256(_mm_loadu_si128(high)),_mm_loadu_si128(low),1);
}

/*
   *sliscp_light192: 18-round sliscp-light permutation
   *state: input state, and output is stored in state again (inplace)
*/
void sliscp_light192_perm_test( u32 *state )
{
        u8 i, j, k;
        u32 t1, t2;
        
        __m256i x2tmp, x3tmp;
        __m256i x[4*PARAL_INST_BY4];
        
        //Packing state
        for ( k = 0; k < PARAL_INST_BY4; k++ )
        {
                x[4*k+0] = _mm256_loadu2_m128i((void *) (state + 4 + 32*k),(void *) (state + 0 + 32*k));
                x[4*k+2] = _mm256_loadu2_m128i((void *) (state + 12 + 32*k),(void *) (state + 8 + 32*k));
                x[4*k+1] = _mm256_loadu2_m128i((void *) (state + 20 + 32*k),(void *) (state + 16 + 32*k));
                x[4*k+3] = _mm256_loadu2_m128i((void *) (state + 28 + 32*k),(void *) (state + 24 + 32*k));
                PACK(x[4*k+0], x[4*k+1], x[4*k+2], x[4*k+3]);
        }
        
        for ( i = 0; i < STEP_ROUND; i++ )
        {
                for ( k = 0; k < PARAL_INST_BY4; k++ )
                {
                        //SSb
                        PACK_SSb(x[4*k+2],x[4*k+3]);
                        for ( j = 0; j < SIMECK_ROUND; j++ )
                        {
                                t1 = (u32)((RC1_192[i] >> j)&1);
                                t2 = (u32)((RC2_192[i] >> j)&1);
                                ROAX(x[4*k+2], x[4*k+3], t1, t2);
                        }
                        UNPACK_SSb(x[4*k+2],x[4*k+3]);
                        //ASc
                        t1 = (u32)SC1_192[i];
                        t2 = (u32)SC2_192[i];
                        x2tmp = x[4*k+0]^SC(t1,t2);
                        x3tmp = x[4*k+1]^SC(t1,t2);
                        //MSb
                        x[4*k+0] = x[4*k+2];
                        x[4*k+1] = x[4*k+3];
                        x2tmp^=x[4*k+2];
                        x3tmp^=x[4*k+3];
                        x[4*k+2] = SWAPBLK(x2tmp);
                        x[4*k+3] = SWAPBLK(x3tmp);
                }
        }
        
        // Unpacking state
        for ( k = 0; k < PARAL_INST_BY4; k++ )
        {
                UNPACK( x[4*k+0], x[4*k+1], x[4*k+2], x[4*k+3] );
                _mm_storeu_si128((void *) (state + 0 + 32*k),_mm256_extracti128_si256(x[4*k+0],0));
                _mm_storeu_si128((void *) (state + 4 + 32*k),_mm256_extracti128_si256(x[4*k+0],1));
                _mm_storeu_si128((void *) (state + 8 + 32*k),_mm256_extracti128_si256(x[4*k+2],0));
                _mm_storeu_si128((void *) (state + 12 + 32*k),_mm256_extracti128_si256(x[4*k+2],1));
                _mm_storeu_si128((void *) (state + 16 + 32*k),_mm256_extracti128_si256(x[4*k+1],0));
                _mm_storeu_si128((void *) (state + 20 + 32*k),_mm256_extracti128_si256(x[4*k+1],1));
                _mm_storeu_si128((void *) (state + 24 + 32*k),_mm256_extracti128_si256(x[4*k+3],0));
                _mm_storeu_si128((void *) (state + 28 + 32*k),_mm256_extracti128_si256(x[4*k+3],1));
        }
        return;
}

/*
   *sliscp_light192: 18-round sliscp-light permutation of width 192 bits
   *x: input state, and output state is stored in state (inplace)
*/
void sliscp_light192( __m256i *x )
{
	u8 i, j, k;
	u32 t1, t2;
        __m256i x2tmp, x3tmp;
        for ( i = 0; i < STEP_ROUND; i++ )
        {
                for ( k = 0; k < PARAL_INST_BY4; k++ )
                {
                        //SSb
                        PACK_SSb(x[4*k+2],x[4*k+3]);
                        for ( j = 0; j < SIMECK_ROUND; j++ )
                        {
                                t1 = (u32)((RC1_192[i] >> j)&1);
                                t2 = (u32)((RC2_192[i] >> j)&1);
                                ROAX(x[4*k+2], x[4*k+3], t1, t2);
                        }
                        UNPACK_SSb(x[4*k+2],x[4*k+3]);
                        //ASc
                        t1 = (u32)SC1_192[i];
                        t2 = (u32)SC2_192[i];
                        x2tmp = x[4*k+0]^SC(t1,t2);
                        x3tmp = x[4*k+1]^SC(t1,t2);
                        //MSb
                        x[4*k+0] = x[4*k+2];
                        x[4*k+1] = x[4*k+3];
                        x2tmp^=x[4*k+2];
                        x3tmp^=x[4*k+3];
                        x[4*k+2] = SWAPBLK(x2tmp);
                        x[4*k+3] = SWAPBLK(x3tmp);
                }
        }
}

/*
   *crypto_aead_encrypt: encrypt plaintext and produce tag
   *key: key 
   *nonce: public nonce
   *adlen: length of ad
   *ad: associate data
   *plen: length of plaintext
   *plaintext: plaintext to be encrypted
   *ciphertext: ciphertext
   *tlen: length of tag (in 32-bit word)
   *tag: tag
*/
void crypto_aead_encrypt( u32 *tag, u32 tlen, u8 *ciphertext, u8 *plaintext, u32 plen, u8 *ad, u32 adlen, u8 *nonce, u8 *key)
{
        u8 k, i;        
	u8 tmp1[8], tmp2[8], tmp3[8], tmp4[8];
	u32 l, mblock, adblock, lmblock, ladblock;
        __m256i x[4*PARAL_INST_BY4];
	u32 index1, index2, index3, index4;

	mblock = plen/8;
	adblock = adlen/8;
	lmblock = plen%8;
	ladblock = adlen%8;
       
	//Initialization
	for ( i = 0; i < PARAL_INST_BY4; i++)
	{
		x[4*i] = _mm256_set_epi8(0x0, nonce[64*i+23], key[64*i+30], key[64*i+31], 0x0, nonce[64*i+20],nonce[64*i+21],nonce[64*i+22], 0x0, nonce[64*i+19], key[64*i+22], key[64*i+23], 0x0, nonce[64*i+16],nonce[64*i+17],nonce[64*i+18], 0x0, nonce[64*i+7], key[64*i+14], key[64*i+15], 0x0, nonce[64*i+4],nonce[64*i+5],nonce[64*i+6], 0x0, nonce[64*i+3], key[64*i+6], key[64*i+7], 0x0, nonce[64*i+0],nonce[64*i+1],nonce[64*i+2]);
		
		x[4*i+1] = _mm256_set_epi8(0x0, nonce[64*i+55], key[64*i+62], key[64*i+63], 0x0, nonce[64*i+52],nonce[64*i+53],nonce[64*i+54], 0x0, nonce[64*i+51], key[64*i+54], key[64*i+55], 0x0, nonce[64*i+48],nonce[64*i+49],nonce[64*i+50], 0x0, nonce[64*i+39], key[64*i+46], key[64*i+47], 0x0, nonce[64*i+36],nonce[64*i+37],nonce[64*i+38], 0x0, nonce[64*i+35], key[64*i+38], key[64*i+39], 0x0, nonce[64*i+32],nonce[64*i+33],nonce[64*i+34]);

		x[4*i+2] = _mm256_set_epi8(0x0, key[64*i+27],key[64*i+28],key[64*i+29], 0x0, key[64*i+24], key[64*i+25], key[64*i+26], 0x0, key[64*i+19], key[64*i+20], key[64*i+21], 0x0, key[64*i+16], key[64*i+17], key[64*i+18], 0x0, key[64*i+11],key[64*i+12],key[64*i+13], 0x0, key[64*i+8], key[64*i+9], key[64*i+10], 0x0, key[64*i+3], key[64*i+4], key[64*i+5], 0x0, key[64*i+0], key[64*i+1], key[64*i+2]);

		x[4*i+3] = _mm256_set_epi8(0x0, key[64*i+59],key[64*i+60],key[64*i+61], 0x0, key[64*i+56], key[64*i+57], key[64*i+58], 0x0, key[64*i+51], key[64*i+52], key[64*i+53], 0x0, key[64*i+48], key[64*i+49], key[64*i+50], 0x0, key[64*i+43],key[64*i+44],key[64*i+45], 0x0, key[64*i+40], key[64*i+41], key[64*i+42], 0x0, key[64*i+35], key[64*i+36], key[64*i+37], 0x0, key[64*i+32], key[64*i+33], key[64*i+34]);

	}
	sliscp_light192(x);
	for ( i = 0; i < PARAL_INST_BY4; i++)
	{
                x[4*i+2] = x[4*i+2]^_mm256_set_epi8(0x0,nonce[64*i+31],0x0,0x0, 0x0,nonce[64*i+28], nonce[64*i+29], nonce[64*i+30], 0x0,nonce[64*i+27],0x0,0x0, 0x0, nonce[64*i+24], nonce[64*i+25], nonce[64*i+26], 0x0, nonce[64*i+15],0x0,0x0, 0x0,nonce[64*i+12], nonce[64*i+13], nonce[64*i+14], 0x0, nonce[64*i+11],0x0,0x0, 0x0, nonce[64*i+8], nonce[64*i+9], nonce[64*i+10]);
                x[4*i+3] = x[4*i+3]^_mm256_set_epi8(0x0,nonce[64*i+63],0x0,0x0, 0x0,nonce[64*i+60], nonce[64*i+61], nonce[64*i+62], 0x0,nonce[64*i+59],0x0,0x0, 0x0, nonce[64*i+56], nonce[64*i+57], nonce[64*i+58], 0x0, nonce[64*i+47],0x0,0x0, 0x0,nonce[64*i+44], nonce[64*i+45], nonce[64*i+46], 0x0, nonce[64*i+43],0x0,0x0, 0x0, nonce[64*i+40], nonce[64*i+41], nonce[64*i+42]);
	}

	//Processing Associated Data
	for ( l = 0; l < adblock; l++ )
	{
		sliscp_light192(x);
		for ( k = 0; k < PARAL_INST_BY4; k++)
		{
			index1 = 8*l + adlen*4*k;
			index2 = 8*l + adlen*4*k+1*adlen;
			index3 = 8*l + adlen*4*k+2*adlen;
			index4 = 8*l + adlen*4*k+3*adlen;
			x[4*k+2] = x[4*k+2]^_mm256_set_epi8(0x0, ad[index2+7], 0x0, 0x0, 0x0, ad[index2+4], ad[index2+5], ad[index2+6], 0x0, ad[index2+3], 0x0, 0x0, 0x0, ad[index2], ad[index2+1], ad[index2+2], 0x0, ad[index1+7], 0x0, 0x0, 0x0, ad[index1+4], ad[index1+5], ad[index1+6], 0x0, ad[index1+3], 0x0, 0x0, 0x0, ad[index1], ad[index1+1], ad[index1+2]);
			x[4*k+3] = x[4*k+3]^_mm256_set_epi8(0x0, ad[index4+7], 0x0, 0x0, 0x0, ad[index4+4], ad[index4+5], ad[index4+6], 0x0, ad[index4+3], 0x0, 0x0, 0x0, ad[index4], ad[index4+1], ad[index4+2], 0x0, ad[index3+7], 0x0, 0x0, 0x0, ad[index3+4], ad[index3+5], ad[index3+6], 0x0, ad[index3+3], 0x0, 0x0, 0x0, ad[index3], ad[index3+1], ad[index3+2]);
        	
			//Domain seperator
			x[4*k] = x[4*k]^_mm256_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x0, 0x0,0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x0, 0x0);
			x[4*k+1] = x[4*k+1]^_mm256_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x0, 0x0,0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x0, 0x0);
		}
	}
	if ( ladblock != 0 )
	{
		sliscp_light192(x);
		for ( k = 0; k < PARAL_INST_BY4; k++)
		{
			index1 = 8*l + adlen*4*k;
			index2 = 8*l + adlen*4*k+1*adlen;
			index3 = 8*l + adlen*4*k+2*adlen;
			index4 = 8*l + adlen*4*k+3*adlen;
			for ( i = 0; i < ladblock; i++ )
			{
				tmp1[i] = ad[index1+i];
				tmp2[i] = ad[index2+i];
				tmp3[i] = ad[index3+i];
				tmp4[i] = ad[index4+i];
			}
			tmp1[ladblock] = 0x80;
			tmp2[ladblock] = 0x80;
			tmp3[ladblock] = 0x80;
			tmp4[ladblock] = 0x80;
			for ( i = ladblock+1; i < 8; i++ )
			{
				tmp1[i] = 0x0;
				tmp2[i] = 0x0;
				tmp3[i] = 0x0;
				tmp4[i] = 0x0;
			}
			x[4*k+2] = x[4*k+2]^_mm256_set_epi8(0x0, tmp2[7], 0x0, 0x0, 0x0, tmp2[4], tmp2[5], tmp2[6], 0x0, tmp2[3], 0x0, 0x0, 0x0, tmp2[0], tmp2[1], tmp2[2], 0x0, tmp1[7], 0x0, 0x0, 0x0, tmp1[4], tmp1[5], tmp1[6], 0x0, tmp1[3], 0x0, 0x0, 0x0, tmp1[0], tmp1[1], tmp1[2]);
			x[4*k+3] = x[4*k+3]^_mm256_set_epi8(0x0, tmp4[7], 0x0, 0x0, 0x0, tmp4[4], tmp4[5], tmp4[6], 0x0, tmp4[3], 0x0, 0x0, 0x0, tmp4[0], tmp4[1], tmp4[2], 0x0, tmp3[7], 0x0, 0x0, 0x0, tmp3[4], tmp3[5], tmp3[6], 0x0, tmp3[3], 0x0, 0x0, 0x0, tmp3[0], tmp3[1], tmp3[2]);
        	
			//Domain seperator
			x[4*k] = x[4*k]^_mm256_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x0, 0x0,0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x0, 0x0);
			x[4*k+1] = x[4*k+1]^_mm256_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x0, 0x0,0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x0, 0x0);
		}
	}
	//Encrypt plaintext
	for ( l = 0; l < mblock; l++ )
	{
		sliscp_light192(x);
		for ( k = 0; k < PARAL_INST_BY4; k++)
		{
			index1 = 8*l + plen*4*k;
			index2 = 8*l + plen*4*k+1*plen;
			index3 = 8*l + plen*4*k+2*plen;
			index4 = 8*l + plen*4*k+3*plen;
			
			ciphertext[index1] = _mm256_extract_epi8 (x[4*k], 2)^plaintext[index1];
			ciphertext[index1+1] = _mm256_extract_epi8 (x[4*k], 1)^plaintext[index1+1];
			ciphertext[index1+2] = _mm256_extract_epi8 (x[4*k], 0)^plaintext[index1+2];
			ciphertext[index1+3] = _mm256_extract_epi8 (x[4*k], 6)^plaintext[index1+3];
			ciphertext[index1+4] = _mm256_extract_epi8 (x[4*k], 10)^plaintext[index1+4];
			ciphertext[index1+5] = _mm256_extract_epi8 (x[4*k], 9)^plaintext[index1+5];
			ciphertext[index1+6] = _mm256_extract_epi8 (x[4*k], 8)^plaintext[index1+6];
			ciphertext[index1+7] = _mm256_extract_epi8 (x[4*k], 14)^plaintext[index1+7];

			ciphertext[index2] = _mm256_extract_epi8 (x[4*k], 18)^plaintext[index2];
			ciphertext[index2+1] = _mm256_extract_epi8 (x[4*k], 17)^plaintext[index2+1];
			ciphertext[index2+2] = _mm256_extract_epi8 (x[4*k], 16)^plaintext[index2+2];
			ciphertext[index2+3] = _mm256_extract_epi8 (x[4*k], 22)^plaintext[index2+3];
			ciphertext[index2+4] = _mm256_extract_epi8 (x[4*k], 26)^plaintext[index2+4];
			ciphertext[index2+5] = _mm256_extract_epi8 (x[4*k], 25)^plaintext[index2+5];
			ciphertext[index2+6] = _mm256_extract_epi8 (x[4*k], 24)^plaintext[index2+6];
			ciphertext[index2+7] = _mm256_extract_epi8 (x[4*k], 30)^plaintext[index2+7];

			x[4*k+2] = x[4*k+2]^_mm256_set_epi8(0x0, plaintext[index2+7], 0x0, 0x0, 0x0, plaintext[index2+4], plaintext[index2+5], plaintext[index2+6], 0x0, plaintext[index2+3], 0x0, 0x0, 0x0, plaintext[index2], plaintext[index2+1], plaintext[index2+2], 0x0, plaintext[index1+7], 0x0, 0x0, 0x0, plaintext[index1+4], plaintext[index1+5], plaintext[index1+6], 0x0, plaintext[index1+3], 0x0, 0x0, 0x0, plaintext[index1], plaintext[index1+1], plaintext[index1+2]);
			ciphertext[index3] = _mm256_extract_epi8 (x[4*k+1], 2)^plaintext[index3];
			ciphertext[index3+1] = _mm256_extract_epi8 (x[4*k+1], 1)^plaintext[index3+1];
			ciphertext[index3+2] = _mm256_extract_epi8 (x[4*k+1], 0)^plaintext[index3+2];
			ciphertext[index3+3] = _mm256_extract_epi8 (x[4*k+1], 6)^plaintext[index3+3];
			ciphertext[index3+4] = _mm256_extract_epi8 (x[4*k+1], 10)^plaintext[index3+4];
			ciphertext[index3+5] = _mm256_extract_epi8 (x[4*k+1], 9)^plaintext[index3+5];
			ciphertext[index3+6] = _mm256_extract_epi8 (x[4*k+1], 8)^plaintext[index3+6];
			ciphertext[index3+7] = _mm256_extract_epi8 (x[4*k+1], 14)^plaintext[index3+7];

			ciphertext[index4] = _mm256_extract_epi8 (x[4*k+1], 18)^plaintext[index4];
			ciphertext[index4+1] = _mm256_extract_epi8 (x[4*k+1], 17)^plaintext[index4+1];
			ciphertext[index4+2] = _mm256_extract_epi8 (x[4*k+1], 16)^plaintext[index4+2];
			ciphertext[index4+3] = _mm256_extract_epi8 (x[4*k+1], 22)^plaintext[index4+3];
			ciphertext[index4+4] = _mm256_extract_epi8 (x[4*k+1], 26)^plaintext[index4+4];
			ciphertext[index4+5] = _mm256_extract_epi8 (x[4*k+1], 25)^plaintext[index4+5];
			ciphertext[index4+6] = _mm256_extract_epi8 (x[4*k+1], 24)^plaintext[index4+6];
			ciphertext[index4+7] = _mm256_extract_epi8 (x[4*k+1], 30)^plaintext[index4+7];
			x[4*k+3] = x[4*k+3]^_mm256_set_epi8(0x0, plaintext[index4+7], 0x0, 0x0, 0x0, plaintext[index4+4], plaintext[index4+5], plaintext[index4+6], 0x0, plaintext[index4+3], 0x0, 0x0, 0x0, plaintext[index4], plaintext[index4+1], plaintext[index4+2], 0x0, plaintext[index3+7], 0x0, 0x0, 0x0, plaintext[index3+4], plaintext[index3+5], plaintext[index3+6], 0x0, plaintext[index3+3], 0x0, 0x0, 0x0, plaintext[index3], plaintext[index3+1], plaintext[index3+2]);
        	
			//Domain seperator
			x[4*k] = x[4*k]^_mm256_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x0, 0x0,0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x0, 0x0);
			x[4*k+1] = x[4*k+1]^_mm256_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x0, 0x0,0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x0, 0x0);
		}
	}
	if ( lmblock != 0 )
	{
		sliscp_light192(x);
		for ( k = 0; k < PARAL_INST_BY4; k++)
		{
			index1 = 8*l + plen*4*k;
			index2 = 8*l + plen*4*k+1*plen;
			index3 = 8*l + plen*4*k+2*plen;
			index4 = 8*l + plen*4*k+3*plen;

			if ( lmblock == 1 )
			{
				ciphertext[index1] = _mm256_extract_epi8 (x[4*k], 2)^plaintext[index1];
				ciphertext[index2] = _mm256_extract_epi8 (x[4*k], 18)^plaintext[index2];

				ciphertext[index3] = _mm256_extract_epi8 (x[4*k+1], 2)^plaintext[index3];
				ciphertext[index4] = _mm256_extract_epi8 (x[4*k+1], 18)^plaintext[index4];
			}
			else if ( lmblock == 2 )
			{
				ciphertext[index1] = _mm256_extract_epi8 (x[4*k], 2)^plaintext[index1];
				ciphertext[index1+1] = _mm256_extract_epi8 (x[4*k], 1)^plaintext[index1+1];
				ciphertext[index2] = _mm256_extract_epi8 (x[4*k], 18)^plaintext[index2];
				ciphertext[index2+1] = _mm256_extract_epi8 (x[4*k], 17)^plaintext[index2+1];

				ciphertext[index3] = _mm256_extract_epi8 (x[4*k+1], 2)^plaintext[index3];
				ciphertext[index3+1] = _mm256_extract_epi8 (x[4*k+1], 1)^plaintext[index3+1];
				ciphertext[index4] = _mm256_extract_epi8 (x[4*k+1], 18)^plaintext[index4];
				ciphertext[index4+1] = _mm256_extract_epi8 (x[4*k+1], 17)^plaintext[index4+1];
			}
			else if ( lmblock == 3 )
			{
				ciphertext[index1] = _mm256_extract_epi8 (x[4*k], 2)^plaintext[index1];
				ciphertext[index1+1] = _mm256_extract_epi8 (x[4*k], 1)^plaintext[index1+1];
				ciphertext[index1+2] = _mm256_extract_epi8 (x[4*k], 0)^plaintext[index1+2];
				ciphertext[index2] = _mm256_extract_epi8 (x[4*k], 18)^plaintext[index2];
				ciphertext[index2+1] = _mm256_extract_epi8 (x[4*k], 17)^plaintext[index2+1];
				ciphertext[index2+2] = _mm256_extract_epi8 (x[4*k], 16)^plaintext[index2+2];

				ciphertext[index3] = _mm256_extract_epi8 (x[4*k+1], 2)^plaintext[index3];
				ciphertext[index3+1] = _mm256_extract_epi8 (x[4*k+1], 1)^plaintext[index3+1];
				ciphertext[index3+2] = _mm256_extract_epi8 (x[4*k+1], 0)^plaintext[index3+2];
				ciphertext[index4] = _mm256_extract_epi8 (x[4*k+1], 18)^plaintext[index4];
				ciphertext[index4+1] = _mm256_extract_epi8 (x[4*k+1], 17)^plaintext[index4+1];
				ciphertext[index4+2] = _mm256_extract_epi8 (x[4*k+1], 16)^plaintext[index4+2];
			}
			else if ( lmblock == 4 )
			{
				ciphertext[index1] = _mm256_extract_epi8 (x[4*k], 2)^plaintext[index1];
				ciphertext[index1+1] = _mm256_extract_epi8 (x[4*k], 1)^plaintext[index1+1];
				ciphertext[index1+2] = _mm256_extract_epi8 (x[4*k], 0)^plaintext[index1+2];
				ciphertext[index1+3] = _mm256_extract_epi8 (x[4*k], 6)^plaintext[index1+3];
				ciphertext[index2] = _mm256_extract_epi8 (x[4*k], 18)^plaintext[index2];
				ciphertext[index2+1] = _mm256_extract_epi8 (x[4*k], 17)^plaintext[index2+1];
				ciphertext[index2+2] = _mm256_extract_epi8 (x[4*k], 16)^plaintext[index2+2];
				ciphertext[index2+3] = _mm256_extract_epi8 (x[4*k], 22)^plaintext[index2+3];

				ciphertext[index3] = _mm256_extract_epi8 (x[4*k+1], 2)^plaintext[index3];
				ciphertext[index3+1] = _mm256_extract_epi8 (x[4*k+1], 1)^plaintext[index3+1];
				ciphertext[index3+2] = _mm256_extract_epi8 (x[4*k+1], 0)^plaintext[index3+2];
				ciphertext[index3+3] = _mm256_extract_epi8 (x[4*k+1], 6)^plaintext[index3+3];
				ciphertext[index4] = _mm256_extract_epi8 (x[4*k+1], 18)^plaintext[index4];
				ciphertext[index4+1] = _mm256_extract_epi8 (x[4*k+1], 17)^plaintext[index4+1];
				ciphertext[index4+2] = _mm256_extract_epi8 (x[4*k+1], 16)^plaintext[index4+2];
				ciphertext[index4+3] = _mm256_extract_epi8 (x[4*k+1], 22)^plaintext[index4+3];
			}
			else if ( lmblock == 5 )
			{
				ciphertext[index1] = _mm256_extract_epi8 (x[4*k], 2)^plaintext[index1];
				ciphertext[index1+1] = _mm256_extract_epi8 (x[4*k], 1)^plaintext[index1+1];
				ciphertext[index1+2] = _mm256_extract_epi8 (x[4*k], 0)^plaintext[index1+2];
				ciphertext[index1+3] = _mm256_extract_epi8 (x[4*k], 6)^plaintext[index1+3];
				ciphertext[index1+4] = _mm256_extract_epi8 (x[4*k], 10)^plaintext[index1+4];
				ciphertext[index2] = _mm256_extract_epi8 (x[4*k], 18)^plaintext[index2];
				ciphertext[index2+1] = _mm256_extract_epi8 (x[4*k], 17)^plaintext[index2+1];
				ciphertext[index2+2] = _mm256_extract_epi8 (x[4*k], 16)^plaintext[index2+2];
				ciphertext[index2+3] = _mm256_extract_epi8 (x[4*k], 22)^plaintext[index2+3];
				ciphertext[index2+4] = _mm256_extract_epi8 (x[4*k], 26)^plaintext[index2+4];

				ciphertext[index3] = _mm256_extract_epi8 (x[4*k+1], 2)^plaintext[index3];
				ciphertext[index3+1] = _mm256_extract_epi8 (x[4*k+1], 1)^plaintext[index3+1];
				ciphertext[index3+2] = _mm256_extract_epi8 (x[4*k+1], 0)^plaintext[index3+2];
				ciphertext[index3+3] = _mm256_extract_epi8 (x[4*k+1], 6)^plaintext[index3+3];
				ciphertext[index3+4] = _mm256_extract_epi8 (x[4*k+1], 10)^plaintext[index3+4];
				ciphertext[index4] = _mm256_extract_epi8 (x[4*k+1], 18)^plaintext[index4];
				ciphertext[index4+1] = _mm256_extract_epi8 (x[4*k+1], 17)^plaintext[index4+1];
				ciphertext[index4+2] = _mm256_extract_epi8 (x[4*k+1], 16)^plaintext[index4+2];
				ciphertext[index4+3] = _mm256_extract_epi8 (x[4*k+1], 22)^plaintext[index4+3];
				ciphertext[index4+4] = _mm256_extract_epi8 (x[4*k+1], 26)^plaintext[index4+4];
			}
			else if ( lmblock == 6 )
			{
				ciphertext[index1] = _mm256_extract_epi8 (x[4*k], 2)^plaintext[index1];
				ciphertext[index1+1] = _mm256_extract_epi8 (x[4*k], 1)^plaintext[index1+1];
				ciphertext[index1+2] = _mm256_extract_epi8 (x[4*k], 0)^plaintext[index1+2];
				ciphertext[index1+3] = _mm256_extract_epi8 (x[4*k], 6)^plaintext[index1+3];
				ciphertext[index1+4] = _mm256_extract_epi8 (x[4*k], 10)^plaintext[index1+4];
				ciphertext[index1+5] = _mm256_extract_epi8 (x[4*k], 9)^plaintext[index1+5];
				ciphertext[index2] = _mm256_extract_epi8 (x[4*k], 18)^plaintext[index2];
				ciphertext[index2+1] = _mm256_extract_epi8 (x[4*k], 17)^plaintext[index2+1];
				ciphertext[index2+2] = _mm256_extract_epi8 (x[4*k], 16)^plaintext[index2+2];
				ciphertext[index2+3] = _mm256_extract_epi8 (x[4*k], 22)^plaintext[index2+3];
				ciphertext[index2+4] = _mm256_extract_epi8 (x[4*k], 26)^plaintext[index2+4];
				ciphertext[index2+5] = _mm256_extract_epi8 (x[4*k], 25)^plaintext[index2+5];

				ciphertext[index3] = _mm256_extract_epi8 (x[4*k+1], 2)^plaintext[index3];
				ciphertext[index3+1] = _mm256_extract_epi8 (x[4*k+1], 1)^plaintext[index3+1];
				ciphertext[index3+2] = _mm256_extract_epi8 (x[4*k+1], 0)^plaintext[index3+2];
				ciphertext[index3+3] = _mm256_extract_epi8 (x[4*k+1], 6)^plaintext[index3+3];
				ciphertext[index3+4] = _mm256_extract_epi8 (x[4*k+1], 10)^plaintext[index3+4];
				ciphertext[index3+5] = _mm256_extract_epi8 (x[4*k+1], 9)^plaintext[index3+5];
				ciphertext[index4] = _mm256_extract_epi8 (x[4*k+1], 18)^plaintext[index4];
				ciphertext[index4+1] = _mm256_extract_epi8 (x[4*k+1], 17)^plaintext[index4+1];
				ciphertext[index4+2] = _mm256_extract_epi8 (x[4*k+1], 16)^plaintext[index4+2];
				ciphertext[index4+3] = _mm256_extract_epi8 (x[4*k+1], 22)^plaintext[index4+3];
				ciphertext[index4+4] = _mm256_extract_epi8 (x[4*k+1], 26)^plaintext[index4+4];
				ciphertext[index4+5] = _mm256_extract_epi8 (x[4*k+1], 25)^plaintext[index4+5];
			}
			else if ( lmblock == 7 )
			{
				ciphertext[index1] = _mm256_extract_epi8 (x[4*k], 2)^plaintext[index1];
				ciphertext[index1+1] = _mm256_extract_epi8 (x[4*k], 1)^plaintext[index1+1];
				ciphertext[index1+2] = _mm256_extract_epi8 (x[4*k], 0)^plaintext[index1+2];
				ciphertext[index1+3] = _mm256_extract_epi8 (x[4*k], 6)^plaintext[index1+3];
				ciphertext[index1+4] = _mm256_extract_epi8 (x[4*k], 10)^plaintext[index1+4];
				ciphertext[index1+5] = _mm256_extract_epi8 (x[4*k], 9)^plaintext[index1+5];
				ciphertext[index1+6] = _mm256_extract_epi8 (x[4*k], 8)^plaintext[index1+6];
				ciphertext[index2] = _mm256_extract_epi8 (x[4*k], 18)^plaintext[index2];
				ciphertext[index2+1] = _mm256_extract_epi8 (x[4*k], 17)^plaintext[index2+1];
				ciphertext[index2+2] = _mm256_extract_epi8 (x[4*k], 16)^plaintext[index2+2];
				ciphertext[index2+3] = _mm256_extract_epi8 (x[4*k], 22)^plaintext[index2+3];
				ciphertext[index2+4] = _mm256_extract_epi8 (x[4*k], 26)^plaintext[index2+4];
				ciphertext[index2+5] = _mm256_extract_epi8 (x[4*k], 25)^plaintext[index2+5];
				ciphertext[index2+6] = _mm256_extract_epi8 (x[4*k], 24)^plaintext[index2+6];

				ciphertext[index3] = _mm256_extract_epi8 (x[4*k+1], 2)^plaintext[index3];
				ciphertext[index3+1] = _mm256_extract_epi8 (x[4*k+1], 1)^plaintext[index3+1];
				ciphertext[index3+2] = _mm256_extract_epi8 (x[4*k+1], 0)^plaintext[index3+2];
				ciphertext[index3+3] = _mm256_extract_epi8 (x[4*k+1], 6)^plaintext[index3+3];
				ciphertext[index3+4] = _mm256_extract_epi8 (x[4*k+1], 10)^plaintext[index3+4];
				ciphertext[index3+5] = _mm256_extract_epi8 (x[4*k+1], 9)^plaintext[index3+5];
				ciphertext[index3+6] = _mm256_extract_epi8 (x[4*k+1], 8)^plaintext[index3+6];
				ciphertext[index4] = _mm256_extract_epi8 (x[4*k+1], 18)^plaintext[index4];
				ciphertext[index4+1] = _mm256_extract_epi8 (x[4*k+1], 17)^plaintext[index4+1];
				ciphertext[index4+2] = _mm256_extract_epi8 (x[4*k+1], 16)^plaintext[index4+2];
				ciphertext[index4+3] = _mm256_extract_epi8 (x[4*k+1], 22)^plaintext[index4+3];
				ciphertext[index4+4] = _mm256_extract_epi8 (x[4*k+1], 26)^plaintext[index4+4];
				ciphertext[index4+5] = _mm256_extract_epi8 (x[4*k+1], 25)^plaintext[index4+5];
				ciphertext[index4+6] = _mm256_extract_epi8 (x[4*k+1], 24)^plaintext[index4+6];
			}

			for ( i = 0; i < lmblock; i++ )
			{
				tmp1[i] = plaintext[index1+i];
				tmp2[i] = plaintext[index2+i];
				tmp3[i] = plaintext[index3+i];
				tmp4[i] = plaintext[index4+i];
			}
			tmp1[lmblock] = 0x80;
			tmp2[lmblock] = 0x80;
			tmp3[lmblock] = 0x80;
			tmp4[lmblock] = 0x80;
			for ( i = lmblock+1; i < 8; i++ )
			{
				tmp1[i] = 0x0;
				tmp2[i] = 0x0;
				tmp3[i] = 0x0;
				tmp4[i] = 0x0;
			}
			x[4*k+2] = x[4*k+2]^_mm256_set_epi8(0x0, tmp2[7], 0x0, 0x0, 0x0, tmp2[4], tmp2[5], tmp2[6], 0x0, tmp2[3], 0x0, 0x0, 0x0, tmp2[0], tmp2[1], tmp2[2], 0x0, tmp1[7], 0x0, 0x0, 0x0, tmp1[4], tmp1[5], tmp1[6], 0x0, tmp1[3], 0x0, 0x0, 0x0, tmp1[0], tmp1[1], tmp1[2]);
			x[4*k+3] = x[4*k+3]^_mm256_set_epi8(0x0, tmp4[7], 0x0, 0x0, 0x0, tmp4[4], tmp4[5], tmp4[6], 0x0, tmp4[3], 0x0, 0x0, 0x0, tmp4[0], tmp4[1], tmp4[2], 0x0, tmp3[7], 0x0, 0x0, 0x0, tmp3[4], tmp3[5], tmp3[6], 0x0, tmp3[3], 0x0, 0x0, 0x0, tmp3[0], tmp3[1], tmp3[2]);
        	
			//Domain seperator
			x[4*k] = x[4*k]^_mm256_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x50, 0x0, 0x0,0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x50, 0x0, 0x0);
			x[4*k+1] = x[4*k+1]^_mm256_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x50, 0x0, 0x0,0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x50, 0x0, 0x0);
		}
	}
	
	//Tag generation phase
	//Extracting tag
	for ( k = 0; k < PARAL_INST_BY4; k++)
	{
        	//Domain seperator
		x[4*k] = x[4*k]^_mm256_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x80, 0x0, 0x0,0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x80, 0x0, 0x0);
		x[4*k+1] = x[4*k+1]^_mm256_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x80, 0x0, 0x0,0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x80, 0x0, 0x0);
	}
	sliscp_light192(x);
	for ( k = 0; k < PARAL_INST_BY4; k++)
	{
                _mm256_storeu2_m128i ((void *) (tag + 4 + 16*k), (void *) (tag + 0 + 16*k), x[4*k+2]);
                _mm256_storeu2_m128i ((void *) (tag + 12 + 16*k), (void *) (tag + 8 + 16*k), x[4*k+3]);
	}
return;
}

/*
   *crypto_aead_decrypt: decrypt ciphertext
   *key: key 
   *nonce: public nonce
   *adlen: length of ad (in byte)
   *ad: associate data
   *tlen: length of tag (in 32-bit word)
   *tag: tag
   *plen: length of plaintext (in byte)
   *plaintext: plaintext to be encrypted
   *ciphertext: ciphertext
*/

void crypto_aead_decrypt( u8 *plaintext,  u8 *ciphertext, u32 plen, u32 *tag, u32 tlen, u8 *ad, u32 adlen, u8 *nonce, u8 *key)
{
        u8 k, i;        
	u8 tmp1[8], tmp2[8], tmp3[8], tmp4[8];
	u32 l, mblock, adblock, lmblock, ladblock;
        __m256i x[4*PARAL_INST_BY4];
	u32 index1, index2, index3, index4;

	mblock = plen/8;
	adblock = adlen/8;
	lmblock = plen%8;
	ladblock = adlen%8;
       
	//Initialization
	for ( i = 0; i < PARAL_INST_BY4; i++)
	{
		x[4*i] = _mm256_set_epi8(0x0, nonce[64*i+23], key[64*i+30], key[64*i+31], 0x0, nonce[64*i+20],nonce[64*i+21],nonce[64*i+22], 0x0, nonce[64*i+19], key[64*i+22], key[64*i+23], 0x0, nonce[64*i+16],nonce[64*i+17],nonce[64*i+18], 0x0, nonce[64*i+7], key[64*i+14], key[64*i+15], 0x0, nonce[64*i+4],nonce[64*i+5],nonce[64*i+6], 0x0, nonce[64*i+3], key[64*i+6], key[64*i+7], 0x0, nonce[64*i+0],nonce[64*i+1],nonce[64*i+2]);
		
		x[4*i+1] = _mm256_set_epi8(0x0, nonce[64*i+55], key[64*i+62], key[64*i+63], 0x0, nonce[64*i+52],nonce[64*i+53],nonce[64*i+54], 0x0, nonce[64*i+51], key[64*i+54], key[64*i+55], 0x0, nonce[64*i+48],nonce[64*i+49],nonce[64*i+50], 0x0, nonce[64*i+39], key[64*i+46], key[64*i+47], 0x0, nonce[64*i+36],nonce[64*i+37],nonce[64*i+38], 0x0, nonce[64*i+35], key[64*i+38], key[64*i+39], 0x0, nonce[64*i+32],nonce[64*i+33],nonce[64*i+34]);

		x[4*i+2] = _mm256_set_epi8(0x0, key[64*i+27],key[64*i+28],key[64*i+29], 0x0, key[64*i+24], key[64*i+25], key[64*i+26], 0x0, key[64*i+19], key[64*i+20], key[64*i+21], 0x0, key[64*i+16], key[64*i+17], key[64*i+18], 0x0, key[64*i+11],key[64*i+12],key[64*i+13], 0x0, key[64*i+8], key[64*i+9], key[64*i+10], 0x0, key[64*i+3], key[64*i+4], key[64*i+5], 0x0, key[64*i+0], key[64*i+1], key[64*i+2]);

		x[4*i+3] = _mm256_set_epi8(0x0, key[64*i+59],key[64*i+60],key[64*i+61], 0x0, key[64*i+56], key[64*i+57], key[64*i+58], 0x0, key[64*i+51], key[64*i+52], key[64*i+53], 0x0, key[64*i+48], key[64*i+49], key[64*i+50], 0x0, key[64*i+43],key[64*i+44],key[64*i+45], 0x0, key[64*i+40], key[64*i+41], key[64*i+42], 0x0, key[64*i+35], key[64*i+36], key[64*i+37], 0x0, key[64*i+32], key[64*i+33], key[64*i+34]);

	}
	sliscp_light192(x);
	for ( i = 0; i < PARAL_INST_BY4; i++)
	{
                x[4*i+2] = x[4*i+2]^_mm256_set_epi8(0x0,nonce[64*i+31],0x0,0x0, 0x0,nonce[64*i+28], nonce[64*i+29], nonce[64*i+30], 0x0,nonce[64*i+27],0x0,0x0, 0x0, nonce[64*i+24], nonce[64*i+25], nonce[64*i+26], 0x0, nonce[64*i+15],0x0,0x0, 0x0,nonce[64*i+12], nonce[64*i+13], nonce[64*i+14], 0x0, nonce[64*i+11],0x0,0x0, 0x0, nonce[64*i+8], nonce[64*i+9], nonce[64*i+10]);
                x[4*i+3] = x[4*i+3]^_mm256_set_epi8(0x0,nonce[64*i+63],0x0,0x0, 0x0,nonce[64*i+60], nonce[64*i+61], nonce[64*i+62], 0x0,nonce[64*i+59],0x0,0x0, 0x0, nonce[64*i+56], nonce[64*i+57], nonce[64*i+58], 0x0, nonce[64*i+47],0x0,0x0, 0x0,nonce[64*i+44], nonce[64*i+45], nonce[64*i+46], 0x0, nonce[64*i+43],0x0,0x0, 0x0, nonce[64*i+40], nonce[64*i+41], nonce[64*i+42]);
	}

	//Processing Associated Data
	for ( l = 0; l < adblock; l++ )
	{
		sliscp_light192(x);
		for ( k = 0; k < PARAL_INST_BY4; k++)
		{
			index1 = 8*l + adlen*4*k;
			index2 = 8*l + adlen*4*k+1*adlen;
			index3 = 8*l + adlen*4*k+2*adlen;
			index4 = 8*l + adlen*4*k+3*adlen;
			x[4*k+2] = x[4*k+2]^_mm256_set_epi8(0x0, ad[index2+7], 0x0, 0x0, 0x0, ad[index2+4], ad[index2+5], ad[index2+6], 0x0, ad[index2+3], 0x0, 0x0, 0x0, ad[index2], ad[index2+1], ad[index2+2], 0x0, ad[index1+7], 0x0, 0x0, 0x0, ad[index1+4], ad[index1+5], ad[index1+6], 0x0, ad[index1+3], 0x0, 0x0, 0x0, ad[index1], ad[index1+1], ad[index1+2]);
			x[4*k+3] = x[4*k+3]^_mm256_set_epi8(0x0, ad[index4+7], 0x0, 0x0, 0x0, ad[index4+4], ad[index4+5], ad[index4+6], 0x0, ad[index4+3], 0x0, 0x0, 0x0, ad[index4], ad[index4+1], ad[index4+2], 0x0, ad[index3+7], 0x0, 0x0, 0x0, ad[index3+4], ad[index3+5], ad[index3+6], 0x0, ad[index3+3], 0x0, 0x0, 0x0, ad[index3], ad[index3+1], ad[index3+2]);
        	
			//Domain seperator
			x[4*k] = x[4*k]^_mm256_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x0, 0x0,0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x0, 0x0);
			x[4*k+1] = x[4*k+1]^_mm256_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x0, 0x0,0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x20, 0x0, 0x0);
		}
	}
	if ( ladblock != 0 )
	{
		sliscp_light192(x);
		for ( k = 0; k < PARAL_INST_BY4; k++)
		{
			index1 = 8*l + adlen*4*k;
			index2 = 8*l + adlen*4*k+1*adlen;
			index3 = 8*l + adlen*4*k+2*adlen;
			index4 = 8*l + adlen*4*k+3*adlen;
			for ( i = 0; i < ladblock; i++ )
			{
				tmp1[i] = ad[index1+i];
				tmp2[i] = ad[index2+i];
				tmp3[i] = ad[index3+i];
				tmp4[i] = ad[index4+i];
			}
			tmp1[ladblock] = 0x80;
			tmp2[ladblock] = 0x80;
			tmp3[ladblock] = 0x80;
			tmp4[ladblock] = 0x80;
			for ( i = ladblock+1; i < 8; i++ )
			{
				tmp1[i] = 0x0;
				tmp2[i] = 0x0;
				tmp3[i] = 0x0;
				tmp4[i] = 0x0;
			}
			x[4*k+2] = x[4*k+2]^_mm256_set_epi8(0x0, tmp2[7], 0x0, 0x0, 0x0, tmp2[4], tmp2[5], tmp2[6], 0x0, tmp2[3], 0x0, 0x0, 0x0, tmp2[0], tmp2[1], tmp2[2], 0x0, tmp1[7], 0x0, 0x0, 0x0, tmp1[4], tmp1[5], tmp1[6], 0x0, tmp1[3], 0x0, 0x0, 0x0, tmp1[0], tmp1[1], tmp1[2]);
			x[4*k+3] = x[4*k+3]^_mm256_set_epi8(0x0, tmp4[7], 0x0, 0x0, 0x0, tmp4[4], tmp4[5], tmp4[6], 0x0, tmp4[3], 0x0, 0x0, 0x0, tmp4[0], tmp4[1], tmp4[2], 0x0, tmp3[7], 0x0, 0x0, 0x0, tmp3[4], tmp3[5], tmp3[6], 0x0, tmp3[3], 0x0, 0x0, 0x0, tmp3[0], tmp3[1], tmp3[2]);
        	
			//Domain seperator
			x[4*k] = x[4*k]^_mm256_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x0, 0x0,0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x0, 0x0);
			x[4*k+1] = x[4*k+1]^_mm256_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x0, 0x0,0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x30, 0x0, 0x0);
		}
	}
	//Encrypt plaintext
	for ( l = 0; l < mblock; l++ )
	{
		sliscp_light192(x);
		for ( k = 0; k < PARAL_INST_BY4; k++)
		{
			index1 = 8*l + plen*4*k;
			index2 = 8*l + plen*4*k+1*plen;
			index3 = 8*l + plen*4*k+2*plen;
			index4 = 8*l + plen*4*k+3*plen;
			plaintext[index1] = _mm256_extract_epi8 (x[4*k], 2)^ciphertext[index1];
			plaintext[index1+1] = _mm256_extract_epi8 (x[4*k], 1)^ciphertext[index1+1];
			plaintext[index1+2] = _mm256_extract_epi8 (x[4*k], 0)^ciphertext[index1+2];
			plaintext[index1+3] = _mm256_extract_epi8 (x[4*k], 6)^ciphertext[index1+3];
			plaintext[index1+4] = _mm256_extract_epi8 (x[4*k], 10)^ciphertext[index1+4];
			plaintext[index1+5] = _mm256_extract_epi8 (x[4*k], 9)^ciphertext[index1+5];
			plaintext[index1+6] = _mm256_extract_epi8 (x[4*k], 8)^ciphertext[index1+6];
			plaintext[index1+7] = _mm256_extract_epi8 (x[4*k], 14)^ciphertext[index1+7];

			plaintext[index2] = _mm256_extract_epi8 (x[4*k], 18)^ciphertext[index2];
			plaintext[index2+1] = _mm256_extract_epi8 (x[4*k], 17)^ciphertext[index2+1];
			plaintext[index2+2] = _mm256_extract_epi8 (x[4*k], 16)^ciphertext[index2+2];
			plaintext[index2+3] = _mm256_extract_epi8 (x[4*k], 22)^ciphertext[index2+3];
			plaintext[index2+4] = _mm256_extract_epi8 (x[4*k], 26)^ciphertext[index2+4];
			plaintext[index2+5] = _mm256_extract_epi8 (x[4*k], 25)^ciphertext[index2+5];
			plaintext[index2+6] = _mm256_extract_epi8 (x[4*k], 24)^ciphertext[index2+6];
			plaintext[index2+7] = _mm256_extract_epi8 (x[4*k], 30)^ciphertext[index2+7];

			x[4*k+2] = x[4*k+2]^_mm256_set_epi8(0x0, plaintext[index2+7], 0x0, 0x0, 0x0, plaintext[index2+4], plaintext[index2+5], plaintext[index2+6], 0x0, plaintext[index2+3], 0x0, 0x0, 0x0, plaintext[index2], plaintext[index2+1], plaintext[index2+2], 0x0, plaintext[index1+7], 0x0, 0x0, 0x0, plaintext[index1+4], plaintext[index1+5], plaintext[index1+6], 0x0, plaintext[index1+3], 0x0, 0x0, 0x0, plaintext[index1], plaintext[index1+1], plaintext[index1+2]);
			plaintext[index3] = _mm256_extract_epi8 (x[4*k+1], 2)^ciphertext[index3];
			plaintext[index3+1] = _mm256_extract_epi8 (x[4*k+1], 1)^ciphertext[index3+1];
			plaintext[index3+2] = _mm256_extract_epi8 (x[4*k+1], 0)^ciphertext[index3+2];
			plaintext[index3+3] = _mm256_extract_epi8 (x[4*k+1], 6)^ciphertext[index3+3];
			plaintext[index3+4] = _mm256_extract_epi8 (x[4*k+1], 10)^ciphertext[index3+4];
			plaintext[index3+5] = _mm256_extract_epi8 (x[4*k+1], 9)^ciphertext[index3+5];
			plaintext[index3+6] = _mm256_extract_epi8 (x[4*k+1], 8)^ciphertext[index3+6];
			plaintext[index3+7] = _mm256_extract_epi8 (x[4*k+1], 14)^ciphertext[index3+7];

			plaintext[index4] = _mm256_extract_epi8 (x[4*k+1], 18)^ciphertext[index4];
			plaintext[index4+1] = _mm256_extract_epi8 (x[4*k+1], 17)^ciphertext[index4+1];
			plaintext[index4+2] = _mm256_extract_epi8 (x[4*k+1], 16)^ciphertext[index4+2];
			plaintext[index4+3] = _mm256_extract_epi8 (x[4*k+1], 22)^ciphertext[index4+3];
			plaintext[index4+4] = _mm256_extract_epi8 (x[4*k+1], 26)^ciphertext[index4+4];
			plaintext[index4+5] = _mm256_extract_epi8 (x[4*k+1], 25)^ciphertext[index4+5];
			plaintext[index4+6] = _mm256_extract_epi8 (x[4*k+1], 24)^ciphertext[index4+6];
			plaintext[index4+7] = _mm256_extract_epi8 (x[4*k+1], 30)^ciphertext[index4+7];
			x[4*k+3] = x[4*k+3]^_mm256_set_epi8(0x0, plaintext[index4+7], 0x0, 0x0, 0x0, plaintext[index4+4], plaintext[index4+5], plaintext[index4+6], 0x0, plaintext[index4+3], 0x0, 0x0, 0x0, plaintext[index4], plaintext[index4+1], plaintext[index4+2], 0x0, plaintext[index3+7], 0x0, 0x0, 0x0, plaintext[index3+4], plaintext[index3+5], plaintext[index3+6], 0x0, plaintext[index3+3], 0x0, 0x0, 0x0, plaintext[index3], plaintext[index3+1], plaintext[index3+2]);
        	
			//Domain seperator
			x[4*k] = x[4*k]^_mm256_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x0, 0x0,0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x0, 0x0);
			x[4*k+1] = x[4*k+1]^_mm256_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x0, 0x0,0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x40, 0x0, 0x0);
		}
	}
	if ( lmblock != 0 )
	{
		sliscp_light192(x);
		for ( k = 0; k < PARAL_INST_BY4; k++)
		{
			index1 = 8*l + plen*4*k;
			index2 = 8*l + plen*4*k+1*plen;
			index3 = 8*l + plen*4*k+2*plen;
			index4 = 8*l + plen*4*k+3*plen;

			if ( lmblock == 1 )
			{
				plaintext[index1] = _mm256_extract_epi8 (x[4*k], 2)^ciphertext[index1];
				plaintext[index2] = _mm256_extract_epi8 (x[4*k], 18)^ciphertext[index2];

				plaintext[index3] = _mm256_extract_epi8 (x[4*k+1], 2)^ciphertext[index3];
				plaintext[index4] = _mm256_extract_epi8 (x[4*k+1], 18)^ciphertext[index4];
			}
			else if ( lmblock == 2 )
			{
				plaintext[index1] = _mm256_extract_epi8 (x[4*k], 2)^ciphertext[index1];
				plaintext[index1+1] = _mm256_extract_epi8 (x[4*k], 1)^ciphertext[index1+1];
				plaintext[index2] = _mm256_extract_epi8 (x[4*k], 18)^ciphertext[index2];
				plaintext[index2+1] = _mm256_extract_epi8 (x[4*k], 17)^ciphertext[index2+1];

				plaintext[index3] = _mm256_extract_epi8 (x[4*k+1], 2)^ciphertext[index3];
				plaintext[index3+1] = _mm256_extract_epi8 (x[4*k+1], 1)^ciphertext[index3+1];
				plaintext[index4] = _mm256_extract_epi8 (x[4*k+1], 18)^ciphertext[index4];
				plaintext[index4+1] = _mm256_extract_epi8 (x[4*k+1], 17)^ciphertext[index4+1];
			}
			else if ( lmblock == 3 )
			{
				plaintext[index1] = _mm256_extract_epi8 (x[4*k], 2)^ciphertext[index1];
				plaintext[index1+1] = _mm256_extract_epi8 (x[4*k], 1)^ciphertext[index1+1];
				plaintext[index1+2] = _mm256_extract_epi8 (x[4*k], 0)^ciphertext[index1+2];
				plaintext[index2] = _mm256_extract_epi8 (x[4*k], 18)^ciphertext[index2];
				plaintext[index2+1] = _mm256_extract_epi8 (x[4*k], 17)^ciphertext[index2+1];
				plaintext[index2+2] = _mm256_extract_epi8 (x[4*k], 16)^ciphertext[index2+2];

				plaintext[index3] = _mm256_extract_epi8 (x[4*k+1], 2)^ciphertext[index3];
				plaintext[index3+1] = _mm256_extract_epi8 (x[4*k+1], 1)^ciphertext[index3+1];
				plaintext[index3+2] = _mm256_extract_epi8 (x[4*k+1], 0)^ciphertext[index3+2];
				plaintext[index4] = _mm256_extract_epi8 (x[4*k+1], 18)^ciphertext[index4];
				plaintext[index4+1] = _mm256_extract_epi8 (x[4*k+1], 17)^ciphertext[index4+1];
				plaintext[index4+2] = _mm256_extract_epi8 (x[4*k+1], 16)^ciphertext[index4+2];
			}
			else if ( lmblock == 4 )
			{
				plaintext[index1] = _mm256_extract_epi8 (x[4*k], 2)^ciphertext[index1];
				plaintext[index1+1] = _mm256_extract_epi8 (x[4*k], 1)^ciphertext[index1+1];
				plaintext[index1+2] = _mm256_extract_epi8 (x[4*k], 0)^ciphertext[index1+2];
				plaintext[index1+3] = _mm256_extract_epi8 (x[4*k], 6)^ciphertext[index1+3];
				plaintext[index2] = _mm256_extract_epi8 (x[4*k], 18)^ciphertext[index2];
				plaintext[index2+1] = _mm256_extract_epi8 (x[4*k], 17)^ciphertext[index2+1];
				plaintext[index2+2] = _mm256_extract_epi8 (x[4*k], 16)^ciphertext[index2+2];
				plaintext[index2+3] = _mm256_extract_epi8 (x[4*k], 22)^ciphertext[index2+3];

				plaintext[index3] = _mm256_extract_epi8 (x[4*k+1], 2)^ciphertext[index3];
				plaintext[index3+1] = _mm256_extract_epi8 (x[4*k+1], 1)^ciphertext[index3+1];
				plaintext[index3+2] = _mm256_extract_epi8 (x[4*k+1], 0)^ciphertext[index3+2];
				plaintext[index3+3] = _mm256_extract_epi8 (x[4*k+1], 6)^ciphertext[index3+3];
				plaintext[index4] = _mm256_extract_epi8 (x[4*k+1], 18)^ciphertext[index4];
				plaintext[index4+1] = _mm256_extract_epi8 (x[4*k+1], 17)^ciphertext[index4+1];
				plaintext[index4+2] = _mm256_extract_epi8 (x[4*k+1], 16)^ciphertext[index4+2];
				plaintext[index4+3] = _mm256_extract_epi8 (x[4*k+1], 22)^ciphertext[index4+3];
			}
			else if ( lmblock == 5 )
			{
				plaintext[index1] = _mm256_extract_epi8 (x[4*k], 2)^ciphertext[index1];
				plaintext[index1+1] = _mm256_extract_epi8 (x[4*k], 1)^ciphertext[index1+1];
				plaintext[index1+2] = _mm256_extract_epi8 (x[4*k], 0)^ciphertext[index1+2];
				plaintext[index1+3] = _mm256_extract_epi8 (x[4*k], 6)^ciphertext[index1+3];
				plaintext[index1+4] = _mm256_extract_epi8 (x[4*k], 10)^ciphertext[index1+4];
				plaintext[index2] = _mm256_extract_epi8 (x[4*k], 18)^ciphertext[index2];
				plaintext[index2+1] = _mm256_extract_epi8 (x[4*k], 17)^ciphertext[index2+1];
				plaintext[index2+2] = _mm256_extract_epi8 (x[4*k], 16)^ciphertext[index2+2];
				plaintext[index2+3] = _mm256_extract_epi8 (x[4*k], 22)^ciphertext[index2+3];
				plaintext[index2+4] = _mm256_extract_epi8 (x[4*k], 26)^ciphertext[index2+4];

				plaintext[index3] = _mm256_extract_epi8 (x[4*k+1], 2)^ciphertext[index3];
				plaintext[index3+1] = _mm256_extract_epi8 (x[4*k+1], 1)^ciphertext[index3+1];
				plaintext[index3+2] = _mm256_extract_epi8 (x[4*k+1], 0)^ciphertext[index3+2];
				plaintext[index3+3] = _mm256_extract_epi8 (x[4*k+1], 6)^ciphertext[index3+3];
				plaintext[index3+4] = _mm256_extract_epi8 (x[4*k+1], 10)^ciphertext[index3+4];
				plaintext[index4] = _mm256_extract_epi8 (x[4*k+1], 18)^ciphertext[index4];
				plaintext[index4+1] = _mm256_extract_epi8 (x[4*k+1], 17)^ciphertext[index4+1];
				plaintext[index4+2] = _mm256_extract_epi8 (x[4*k+1], 16)^ciphertext[index4+2];
				plaintext[index4+3] = _mm256_extract_epi8 (x[4*k+1], 22)^ciphertext[index4+3];
				plaintext[index4+4] = _mm256_extract_epi8 (x[4*k+1], 26)^ciphertext[index4+4];
			}
			else if ( lmblock == 6 )
			{
				plaintext[index1] = _mm256_extract_epi8 (x[4*k], 2)^ciphertext[index1];
				plaintext[index1+1] = _mm256_extract_epi8 (x[4*k], 1)^ciphertext[index1+1];
				plaintext[index1+2] = _mm256_extract_epi8 (x[4*k], 0)^ciphertext[index1+2];
				plaintext[index1+3] = _mm256_extract_epi8 (x[4*k], 6)^ciphertext[index1+3];
				plaintext[index1+4] = _mm256_extract_epi8 (x[4*k], 10)^ciphertext[index1+4];
				plaintext[index1+5] = _mm256_extract_epi8 (x[4*k], 9)^ciphertext[index1+5];
				plaintext[index2] = _mm256_extract_epi8 (x[4*k], 18)^ciphertext[index2];
				plaintext[index2+1] = _mm256_extract_epi8 (x[4*k], 17)^ciphertext[index2+1];
				plaintext[index2+2] = _mm256_extract_epi8 (x[4*k], 16)^ciphertext[index2+2];
				plaintext[index2+3] = _mm256_extract_epi8 (x[4*k], 22)^ciphertext[index2+3];
				plaintext[index2+4] = _mm256_extract_epi8 (x[4*k], 26)^ciphertext[index2+4];
				plaintext[index2+5] = _mm256_extract_epi8 (x[4*k], 25)^ciphertext[index2+5];

				plaintext[index3] = _mm256_extract_epi8 (x[4*k+1], 2)^ciphertext[index3];
				plaintext[index3+1] = _mm256_extract_epi8 (x[4*k+1], 1)^ciphertext[index3+1];
				plaintext[index3+2] = _mm256_extract_epi8 (x[4*k+1], 0)^ciphertext[index3+2];
				plaintext[index3+3] = _mm256_extract_epi8 (x[4*k+1], 6)^ciphertext[index3+3];
				plaintext[index3+4] = _mm256_extract_epi8 (x[4*k+1], 10)^ciphertext[index3+4];
				plaintext[index3+5] = _mm256_extract_epi8 (x[4*k+1], 9)^ciphertext[index3+5];
				plaintext[index4] = _mm256_extract_epi8 (x[4*k+1], 18)^ciphertext[index4];
				plaintext[index4+1] = _mm256_extract_epi8 (x[4*k+1], 17)^ciphertext[index4+1];
				plaintext[index4+2] = _mm256_extract_epi8 (x[4*k+1], 16)^ciphertext[index4+2];
				plaintext[index4+3] = _mm256_extract_epi8 (x[4*k+1], 22)^ciphertext[index4+3];
				plaintext[index4+4] = _mm256_extract_epi8 (x[4*k+1], 26)^ciphertext[index4+4];
				plaintext[index4+5] = _mm256_extract_epi8 (x[4*k+1], 25)^ciphertext[index4+5];
			}
			else if ( lmblock == 7 )
			{
				plaintext[index1] = _mm256_extract_epi8 (x[4*k], 2)^ciphertext[index1];
				plaintext[index1+1] = _mm256_extract_epi8 (x[4*k], 1)^ciphertext[index1+1];
				plaintext[index1+2] = _mm256_extract_epi8 (x[4*k], 0)^ciphertext[index1+2];
				plaintext[index1+3] = _mm256_extract_epi8 (x[4*k], 6)^ciphertext[index1+3];
				plaintext[index1+4] = _mm256_extract_epi8 (x[4*k], 10)^ciphertext[index1+4];
				plaintext[index1+5] = _mm256_extract_epi8 (x[4*k], 9)^ciphertext[index1+5];
				plaintext[index1+6] = _mm256_extract_epi8 (x[4*k], 8)^ciphertext[index1+6];
				plaintext[index2] = _mm256_extract_epi8 (x[4*k], 18)^ciphertext[index2];
				plaintext[index2+1] = _mm256_extract_epi8 (x[4*k], 17)^ciphertext[index2+1];
				plaintext[index2+2] = _mm256_extract_epi8 (x[4*k], 16)^ciphertext[index2+2];
				plaintext[index2+3] = _mm256_extract_epi8 (x[4*k], 22)^ciphertext[index2+3];
				plaintext[index2+4] = _mm256_extract_epi8 (x[4*k], 26)^ciphertext[index2+4];
				plaintext[index2+5] = _mm256_extract_epi8 (x[4*k], 25)^ciphertext[index2+5];
				plaintext[index2+6] = _mm256_extract_epi8 (x[4*k], 24)^ciphertext[index2+6];

				plaintext[index3] = _mm256_extract_epi8 (x[4*k+1], 2)^ciphertext[index3];
				plaintext[index3+1] = _mm256_extract_epi8 (x[4*k+1], 1)^ciphertext[index3+1];
				plaintext[index3+2] = _mm256_extract_epi8 (x[4*k+1], 0)^ciphertext[index3+2];
				plaintext[index3+3] = _mm256_extract_epi8 (x[4*k+1], 6)^ciphertext[index3+3];
				plaintext[index3+4] = _mm256_extract_epi8 (x[4*k+1], 10)^ciphertext[index3+4];
				plaintext[index3+5] = _mm256_extract_epi8 (x[4*k+1], 9)^ciphertext[index3+5];
				plaintext[index3+6] = _mm256_extract_epi8 (x[4*k+1], 8)^ciphertext[index3+6];
				plaintext[index4] = _mm256_extract_epi8 (x[4*k+1], 18)^ciphertext[index4];
				plaintext[index4+1] = _mm256_extract_epi8 (x[4*k+1], 17)^ciphertext[index4+1];
				plaintext[index4+2] = _mm256_extract_epi8 (x[4*k+1], 16)^ciphertext[index4+2];
				plaintext[index4+3] = _mm256_extract_epi8 (x[4*k+1], 22)^ciphertext[index4+3];
				plaintext[index4+4] = _mm256_extract_epi8 (x[4*k+1], 26)^ciphertext[index4+4];
				plaintext[index4+5] = _mm256_extract_epi8 (x[4*k+1], 25)^ciphertext[index4+5];
				plaintext[index4+6] = _mm256_extract_epi8 (x[4*k+1], 24)^ciphertext[index4+6];
			}
			for ( i = 0; i < lmblock; i++ )
			{
				tmp1[i] = plaintext[index1+i];
				tmp2[i] = plaintext[index2+i];
				tmp3[i] = plaintext[index3+i];
				tmp4[i] = plaintext[index4+i];
			}
			tmp1[lmblock] = 0x80;
			tmp2[lmblock] = 0x80;
			tmp3[lmblock] = 0x80;
			tmp4[lmblock] = 0x80;
			for ( i = lmblock+1; i < 8; i++ )
			{
				tmp1[i] = 0x0;
				tmp2[i] = 0x0;
				tmp3[i] = 0x0;
				tmp4[i] = 0x0;
			}
			x[4*k+2] = x[4*k+2]^_mm256_set_epi8(0x0, tmp2[7], 0x0, 0x0, 0x0, tmp2[4], tmp2[5], tmp2[6], 0x0, tmp2[3], 0x0, 0x0, 0x0, tmp2[0], tmp2[1], tmp2[2], 0x0, tmp1[7], 0x0, 0x0, 0x0, tmp1[4], tmp1[5], tmp1[6], 0x0, tmp1[3], 0x0, 0x0, 0x0, tmp1[0], tmp1[1], tmp1[2]);
			x[4*k+3] = x[4*k+3]^_mm256_set_epi8(0x0, tmp4[7], 0x0, 0x0, 0x0, tmp4[4], tmp4[5], tmp4[6], 0x0, tmp4[3], 0x0, 0x0, 0x0, tmp4[0], tmp4[1], tmp4[2], 0x0, tmp3[7], 0x0, 0x0, 0x0, tmp3[4], tmp3[5], tmp3[6], 0x0, tmp3[3], 0x0, 0x0, 0x0, tmp3[0], tmp3[1], tmp3[2]);
        	
			//Domain seperator
			x[4*k] = x[4*k]^_mm256_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x50, 0x0, 0x0,0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x50, 0x0, 0x0);
			x[4*k+1] = x[4*k+1]^_mm256_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x50, 0x0, 0x0,0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x50, 0x0, 0x0);
		}
	}
	
	//Tag generation phase
	//Extracting tag
	for ( k = 0; k < PARAL_INST_BY4; k++)
	{
        	//Domain seperator
		x[4*k] = x[4*k]^_mm256_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x80, 0x0, 0x0,0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x80, 0x0, 0x0);
		x[4*k+1] = x[4*k+1]^_mm256_set_epi8(0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x80, 0x0, 0x0,0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x0, 0x80, 0x0, 0x0);
	}
	sliscp_light192(x);
	for ( k = 0; k < PARAL_INST_BY4; k++)
	{
                _mm256_storeu2_m128i ((void *) (tag + 4 + 16*k), (void *) (tag + 0 + 16*k), x[4*k+2]);
                _mm256_storeu2_m128i ((void *) (tag + 12 + 16*k), (void *) (tag + 8 + 16*k), x[4*k+3]);
	}
return;
}
